/*********************************************************************
 *		  
 * Filename:	  irport.c
 * Version:	  0.1
 * Description:  THis does the main higher level functions of the driver (ie.
 * file operations, etc.
 * Status:	  Experimental.
 * Author:	  Ryan Shillington
 *
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of
 *    the License, or (at your option) any later version.
 *
 ********************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/errno.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/serial_reg.h>
#include <linux/delay.h>

#include <linux/fs.h>

#include "driver.h"
#include "irport.h"
#include "crc.h"
#include "wrapper.h"
#include "ircompat.h"

#define IO_EXTENT	8

/* Define the file stuff */
int file_read(struct inode *inodeP, struct file *fileP, 
    char *buffP, int lenI);
int file_write(struct inode *inodeP, struct file *fileP, 
    const char *buffP, int lenI);

static struct file_operations fops = {
  NULL,  /* seek */
  file_read,  /* read */
  file_write,
  NULL,   /* readdir */
  NULL,
  NULL,   /* ioctl */
  NULL,   /* mmap */
  NULL, /* open */
  NULL, /* close */
  NULL,
  NULL,
};

/* How long before we timeout */
#define MEDIABUSY_TIMEOUT 1000

static void irport_interrupt(int irq, void *dev_id, struct pt_regs *regs);

int irport_is_receiving( struct driver_cb *self);
void irport_start_media_timer( struct driver_cb *driver);
int irport_is_tbusy(struct driver_cb *driver);
void irport_media_busy_expired( unsigned long data);
void irport_todo_expired( unsigned long data);
void irport_start_todo_timer( struct driver_cb *driver, int timeout);
void irport_async_unwrap( struct driver_cb *driver, __u8 byte);
int irport_async_wrap( struct sk_buff *skb, __u8 *tx_buff, int buffsize);

static char *driver_name = "irport_w83977af";
struct driver_cb *irport_drv;

/*
 * Initialize IO port 
 */
int irport_init( void) 
{
	struct driver_cb *driver;
  int result;
        
	DEBUG(4, "irport_init - START\n");

  /* 1 - Initialize the driver */
	driver = kmalloc(sizeof(struct driver_cb), GFP_KERNEL);
	memset(driver, 0, sizeof(struct driver_cb));

  /* 2 - Keep track of the driver (put it in a global variable) */
	irport_drv = driver;

  /* 3 - Set the specific information about the hardware */
	driver->baud = 9600;
  driver->driver_name = driver_name;
	driver->SCE.iobase = 0x400;
	driver->SCE.irq = 22;
	driver->SCE.ioextent = 8;
	driver->SCE.dma = 0;
	driver->hard_xmit = irport_xmit;
	driver->irq_handler = irport_interrupt;
	driver->change_speed = irport_setspeed;
	driver->is_media_busy = irport_is_media_busy;
	driver->set_media_busy = irport_set_media_busy;
	driver->is_receiving = irport_is_receiving;
	driver->is_tbusy = irport_is_tbusy;
	driver->data = NULL;
  driver->rx.state = OUTSIDE_FRAME;
  driver->readq = NULL;

  /* 4 - Allocate and initialize memory for Transmit and Receive */
  driver->rx.buff = ( __u8 *) kmalloc( BUF_MAX_LEN, GFP_KERNEL|GFP_DMA);
  driver->tx.buff = ( __u8 *) kmalloc( BUF_MAX_LEN, GFP_KERNEL|GFP_DMA);
  memset(driver->rx.buff, 0, BUF_MAX_LEN);
  memset(driver->tx.buff, 0, BUF_MAX_LEN);
  driver->rx.buffsize=BUF_MAX_LEN;
  driver->tx.buffsize=BUF_MAX_LEN;
  skb_queue_head_init( &driver->rx.list);
  skb_queue_head_init( &driver->tx.list);

  /* 5 - Detect to make sure that the device exists */
	if (FIR_detect(driver) < 0) {
		DEBUG(0, "Error, device not found!\n");
		return -1;
	}
  
  /* 6 - Lock the port that we need */
  if (check_region(driver->SCE.iobase, driver->SCE.ioextent) < 0) 
  { 
    DEBUG(0, "IrVDD: can't get SCE iobase of 0x%03x\n", driver->SCE.iobase);
    irport_cleanup(driver); 
    return -ENODEV;
  } else {
    request_region( driver->SCE.iobase,
                    driver->SCE.ioextent,
                    driver->driver_name);
  }

  /* 7 - Register the IRQ handler */
  if(driver->SCE.irq > 0)
  {
    request_irq(driver->SCE.irq, driver->irq_handler, SA_INTERRUPT,
      driver->driver_name, (void *) driver);
  }

  /* 8 - The device is initially NOT busy */
	irport_set_media_busy(driver, FALSE); 

  /* 9 - I don't know what this does that FIR_detect doesn't do */
	irport_start(driver);

  /* 10 - Set the speed of the IR device */
	irport_setspeed( driver, 9600 ); 

  /* 11 - Set up device and register it */
  driver->fops = &fops;
  driver->driver_id = 60;
  result = register_chrdev(driver->driver_id, driver_name, driver->fops);
  if (result < 0) {
    DEBUG(0, "IRPORT: Can't get major device number\n");
    return result;
  }

  /* 12 - Print out a message to confirm that this function is finished */
	DEBUG(4, "irport_init - FINISHED\n");

	return 0;
}

int file_read(struct inode *inodeP, struct file *fileP, 
    char *buffP, int lenI)
{
	struct sk_buff *skb;
  int totalI = 0;   /* Amount of data copied into buffP */

  DEBUG(3, "Read data...\n");

  //* 1 - Block if there is no data waiting to be read
  while (skb_queue_empty(&(irport_drv->rx.list)))
  {
    if (fileP->f_flags & O_NONBLOCK)
      return -EWOULDBLOCK;
    DEBUG(3, "Blocking, waiting for data to come in.\n");
    interruptible_sleep_on(&irport_drv->readq);
    if (current->signal & ~current->blocked) //* A signal arrived
      return -ERESTARTSYS;                  //* tell the fs layer to handle it
  }

  /* 2 - Get data received */
  /* 2.1 - Find the information in the first packet */
  skb = skb_peek((&irport_drv->rx.list));

  /* 2.2 - Swallow as many whole packets as possible */
  while ((!skb_queue_empty(&(irport_drv->rx.list))) && (skb->len < (lenI-totalI)))
  {
    /*2.3 Take the packet off of the cue */
    skb = skb_dequeue(&(irport_drv->rx.list));

    /*2.4 Copy the data out */
    memcpy(buffP+totalI, skb->data, skb->len);
    totalI += skb->len;
    
    /*2.5 Throw away the packaging */
    dev_kfree_skb(skb, FREE_READ);

    /*2.6 Get the next packet */
    skb = skb_peek(&(irport_drv->rx.list));
  }
  
  /* 2.7 Get a partial amount of data from the packet, if we can't finish it
   * off */
  if(!skb_queue_empty(&(irport_drv->rx.list)))
  {
    memcpy(buffP+totalI, skb->data, lenI-totalI);
    skb_pull(skb, lenI-totalI);
    totalI = lenI;
  }
  
  DEBUG(3, "Read successful.  %d bytes read\n", totalI);
  return totalI;
}

int file_write(struct inode *inodep, struct file *filep, 
    const char *buffP, int lenI)
{
	struct sk_buff *skb;

  DEBUG(3, "Writing data...\n");

  /* 1 - Sanity Check */
  if (lenI < 0)
    return 0;

  /* 2 - Allocate memory */
  skb = alloc_skb(lenI+2, 1);
  if(!skb) {
    DEBUG(0, "IRDA: TX PROBLEMS!!! OUT OF MEMORY!!!\n");
    return 0;
  }
    
  /* 3 - Put data into the skbuff */
  memcpy(skb_put(skb, lenI), buffP, lenI);
  skb->ip_summed = CHECKSUM_UNNECESSARY;    /* Don't check */

  /* 4 - Bombs away!!! Put the data on the list of things to get shipped
   * out! */
  irport_xmit(irport_drv, skb);
  
  return lenI;
}

/*
 * Function irport_cleanup ()
 *
 *    irport_cleanup(): Removes everything!
 *
 */
void irport_cleanup(struct driver_cb *driver) 
{
	int SCE_iobase, result;
  struct sk_buff *skb;

	DEBUG(4, "irport_cleanup - START\n");

  /* 1 - Initialize important variables */
	SCE_iobase = driver->SCE.iobase;

  /* 2 - Shutdown the device */
	FIR_shutdown(driver);

	/* 3 - Reset uart */
	outb_p( 0, SCE_iobase+UART_MCR);

	/* 4 - Turn off interrupts */
  DEBUG(4, "Turning off interupts\n");
	outb_p( 0, SCE_iobase+UART_IER); 

  /* 5 - Unregister the interrupts */
  DEBUG(4, "Releasing IRQ %d\n", driver->SCE.irq);
  if(driver->SCE.irq >0)
    free_irq(driver->SCE.irq, driver);

  /* 6 - Release the PORT that this driver is using */
  DEBUG(4, "Releasing Region %03x\n", driver->SCE.iobase);
  release_region( driver->SCE.iobase, driver->SCE.ioextent);

	/* 7 - Stop media busy timer */
	del_timer( &driver->media_busy_timer);
	del_timer( &driver->todo_timer);

  /* 8 - Empty tx_list and rx_list! */
  DEBUG(4, "Releasing transmit and receive buffers \n");
  if ( driver->tx.buff) {
    kfree_s( driver->tx.buff, driver->tx.buffsize);
  }
  if ( driver->rx.buff) {
    kfree_s( driver->rx.buff, driver->rx.buffsize);
  }
  while (( skb = skb_dequeue( &driver->rx.list)) != NULL) {
    DEV_KFREE_SKB( skb, FREE_READ);
  }
  while (( skb = skb_dequeue( &driver->tx.list)) != NULL) {
    DEV_KFREE_SKB( skb, FREE_WRITE);
  }
  
  /* 9 - Unregister this driver's major number */
  result = unregister_chrdev(driver->driver_id, driver_name);
  if (result < 0)
  {
    DEBUG(0, "ERROR: Could not deregister irda device(Major #%d).  Received %d\n",
        driver->driver_id, result);
  }

  /* 10 - Delete the driver out of memory (This must be last) */
  if (driver)
        kfree_s(driver, sizeof(struct driver_cb));

}

/*
 * Function irport_start (void)
 *
 *    Start IO port 
 *
 */
int irport_start(struct driver_cb *driver) 
{
	int iobase = driver->SCE.iobase;

  /*
   * Reset the UART.
   */

  outb_p( 0x00, iobase+UART_MCR);
  outb_p((UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT), iobase+UART_FCR);
  (void)inb( iobase+UART_RX);

	/* Initialize UART */
	outb_p( UART_LCR_WLEN8, iobase+UART_LCR);  /* Reset DLAB */
	outb_p(( UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2), iobase+UART_MCR);

	/* Turn on interrups */
	outb_p(( UART_IER_THRI |UART_IER_RLSI | UART_IER_RDI), iobase+UART_IER); 
	return 0;
}

/*
 * Function irport_setspeed (speed)
 *
 *    Set speed of port to specified baudrate
 *
 */
void irport_setspeed( struct driver_cb *driver, int speed) 
{
	int iobase = driver->SCE.iobase;
	int fcr;    /* FIFO control reg */
	int lcr;    /* Line control reg */
	int divisor;

	DEBUG(0, "%s: Setting speed to: %d\n", driver->driver_name, speed);

	if (speed > 115200) {
		FIR_setspeed(driver, speed);
	} else {
		divisor = SPEED_MAX / speed;
	  
		fcr = UART_FCR_ENABLE_FIFO | UART_FCR_TRIGGER_14;
        
		/* IrDA ports use 8N1 */
		lcr = UART_LCR_WLEN8;

		outb_p( UART_LCR_DLAB | lcr, iobase+UART_LCR); /* Set DLAB */
		outb_p( divisor & 0xff,      iobase+UART_DLL); /* Set speed	*/
		outb_p( divisor >> 8,	     iobase+UART_DLM);
		outb_p( lcr,		     iobase+UART_LCR); /* Set 8N1	*/
		outb_p( fcr,		     iobase+UART_FCR); /* Enable FIFO's */
	}

	driver->baud = speed;

}

/*
 * Function irport_interrupt (irq, dev_id, regs)
 *
 *    
 */
static void irport_interrupt(int irq, void *dev_id, struct pt_regs *regs) 
{
	struct driver_cb *driver = (struct driver_cb *) dev_id;

	int iobase, status;
	int iir;

	DEBUG(4, "irport_interrupt: irq %d\n", irq);

	if (!driver) {
		printk(KERN_WARNING "irport_interrupt: irq %d for unknown device.\n", irq);
		return;
	}
	
	if (driver->baud > 115200) {
		FIR_interrupt(irq, dev_id, regs);
	} else {
		iobase = driver->SCE.iobase;
		iir = inb( iobase+UART_IIR);
		do {
			status = inb( iobase+UART_LSR);

			if ( status & UART_LSR_DR) {
				/* Receive interrupt */
        DEBUG(4, "Received Data\n");
				irport_receive(driver);
			}
			if ( status & UART_LSR_THRE) {
				/* Transmitter ready for data */
        DEBUG(4, "Ready to Transmit Data\n");
				irport_write(driver);
			}
		} while ( !(inb( iobase+UART_IIR) & UART_IIR_NO_INT));
	}

	return;

}

int irport_write(struct driver_cb *driver)
{
	int iobase = driver->SCE.iobase;
	int fifo_size, cnt, count=0;
	struct sk_buff *skb;
  static int writeNum = 0;
	
  writeNum++;
	if (!(inb_p( iobase+UART_LSR) & UART_LSR_THRE)) {
		DEBUG(0, "irport_write failed, fifo not empty!\n");
		return 0;
	}
        
	fifo_size = driver->SCE.fifo_size - 2;

	/* Transmit current frame */	        
	do {
/* Avoid failed transmit byte in 16550A, if FIFO get empty and only one byte 
 * left to transmit, then that byte may not be transmited until the next frame
 * fifo_size = 2 means at this point that there is only 1 place left in fifo. 
 * We break here, so that there are 2 bytes left in buffer to transmit at next
 * interrupt
 */
		cnt = driver->tx.count - driver->tx.ptr;
		if ((cnt == 2) && (fifo_size == 2)) {
			break;
		}
		if (cnt <= 0) {
			DEBUG(4, "irport_write: finished with frame\n"); 
			DEBUG(4, "queue len: %d\n", skb_queue_len( &driver->tx.list));
			if ((skb = skb_dequeue(&driver->tx.list))) {
				irport_xmit( driver, skb);
			}
			break;
		}	        

		/* Transmit next byte */
		outb( *driver->tx.head, iobase+UART_TX);
    //printk("%x", *driver->tx.head);
	  DEBUG(5, "%c", (char) *((char *)driver->tx.head));
		driver->tx.ptr++;
		driver->tx.head++;
		count++;
	} while (--fifo_size > 0);
	  DEBUG(4, "Write finished.  %d bytes written\n", count);
       
	return count;
}

/*
 * Function irport_xmit (void)
 *
 *    Transmits the current frame until FIFO is full, then
 *    waits until the next transmitt interrupt, and continues until the
 *    frame is transmited.
 */

void irport_xmit(struct driver_cb *driver, struct sk_buff *skb) 
{
	DEBUG(4, "irport_xmit()\n");

	if ( !driver)
		return;
	        
	if (driver->baud > 115200) {
		DEBUG(4, "irport_xmit: baud=%d, sending to FIR_xmit!\n", driver->baud);
		DEBUG(4, "irport_xmit: skb->len=%ld\n", skb->len);
		memcpy(driver->tx.buff, skb->data, skb->len);
		driver->tx.count = skb->len;
		driver->tx.head = driver->tx.buff;
		driver->tx.ptr = 0;
		FIR_xmit(driver);
	} else {
		driver->tx.count = irport_async_wrap( skb, 
						      driver->tx.buff, driver->tx.buffsize);
		driver->tx.head = driver->tx.buff;
		driver->tx.ptr = 0;

		if(irport_write(driver) < 0)
    {
      skb_queue_tail(&(irport_drv->tx.list), skb);
      return;
    }

	}


	IS_SKB( skb, return;);
	FREE_SKB_MAGIC( skb);
	DEV_KFREE_SKB( skb, FREE_WRITE);
}
        
/*
 * Function irport_receive (void)
 *
 *    Reads available bytes from the FIFO, and unwraps into 
 *    frame(s), which are eventually passed up to the upper layer in the
 *    protocol stack 
 *
 *    return: number of bytes received (may not be entire frame)
 *	 	-1 on error
 */
int irport_receive(struct driver_cb *driver) 
{
	__u8 byte = 0x00;
	int iobase;
	int count = 0;

	if ( !driver)
		return -1;

	DEBUG(4, "--> irport_receive\n");
	
	iobase = driver->SCE.iobase;

	/* 
	 *  Receive all characters available in FIFO 
	 */
	do {
		byte = inb_p( iobase+UART_RX);
		irport_async_unwrap(driver, byte);
		count++;
    printk("%c", ((char) byte));
	} while ( inb_p( iobase+UART_LSR) & UART_LSR_DR);

	DEBUG(4, "irport_receive: got %d bytes\n", count);
 
	return count;
}

/*
 * Function irport_set_media_busy (status)
 *
 *    Called when we have detected that another station is transmitting
 *    in contention mode.
 *
 */
void irport_set_media_busy( struct driver_cb *driver, int status) 
{
	if (driver == NULL)
		return;

	if ( status) {
		driver->media_busy = TRUE;
		irport_start_media_timer(driver);
		DEBUG(4, "Media busy!\n");
	} else {
		driver->media_busy = FALSE;
		del_timer( &driver->media_busy_timer);
	}
}

void irport_start_media_timer( struct driver_cb *driver) 
{
	del_timer( &driver->media_busy_timer);
        
	driver->media_busy_timer.data	  = (unsigned long) driver;
	driver->media_busy_timer.function = &irport_media_busy_expired;
	driver->media_busy_timer.expires  = jiffies + MEDIABUSY_TIMEOUT;
        
	add_timer( &driver->media_busy_timer);
}


/*
 * Function irtty_media_busy_expired (data)
 *
 *    
 */
void irport_media_busy_expired(unsigned long data) 
{
	struct driver_cb *driver = (struct driver_cb *) data;

	DEBUG(4, "Mediabusy timer expired!\n");
        
	irport_set_media_busy( driver, FALSE);
}

void irport_start_todo_timer( struct driver_cb *driver, int timeout) 
{
	del_timer( &driver->todo_timer);
        
	driver->todo_timer.data     = (unsigned long) driver;
	driver->todo_timer.function = &irport_todo_expired;
	driver->todo_timer.expires  = jiffies + timeout;
        
	add_timer( &driver->todo_timer);
}

void irport_todo_expired( unsigned long data)
{
	DEBUG(4, "Todo timer expired!\n");
}

int irport_is_media_busy(struct driver_cb *driver) 
{
	return driver->media_busy;
}

int irport_is_tbusy(struct driver_cb *driver) 
{
	return driver->media_busy;
}

/*
 * Function irport_is_receiving (irvdd)
 *
 */
int irport_is_receiving( struct driver_cb *self)
{
	DEBUG(4, "IrPORT: is_receiving()\n");

	if (!self || self->magic != DRIVER_CB_MAGIC) {
		DEBUG(0, "irport_is_receiving: BAD MAGIC!\n");
		return FALSE;
	}

	return 1;
}

/*
 * Function async_unwrap (skb)
 *
 *    Parse and de-stuff frame received from the IR-port
 *
 */

void irport_async_unwrap( struct driver_cb *driver, __u8 byte) 
{
	int state = driver->rx.state;
	struct sk_buff *skb;

	DEBUG( 4, "async_unwrap() State = %d\n", state);

	/* State machine for receiving frames */	   
	switch( state) {
	case OUTSIDE_FRAME:
		if ( byte == BOF) {
			DEBUG(4, "BOF found, setting state to BEGIN_FRAME and in_frame to TRUE\n");
			state = BEGIN_FRAME;
			driver->rx.in_frame = TRUE;
			driver->rx.count = 0; 
		} else if ( byte == EOF) {
      DEBUG(4, "Received EOF in OUTSIDE_FRAME state\n");
			irport_set_media_busy( driver, TRUE);
		}
		break;
	case BEGIN_FRAME:
		switch ( byte) {
		case BOF:
			DEBUG(4, "XBOF found!\n");
			/* Continue */
			break;
		case CE:
			DEBUG(4, "CE found!\n");
			/* Stuffed byte */
			state = LINK_ESCAPE;
			break;
		case EOF:
			/* Abort frame */
			DEBUG(4, "Frame abort (1)\n");
			state = OUTSIDE_FRAME;
			break;
		default:
			/* Got first byte of frame */
			if ( driver->rx.count < driver->rx.buffsize)  {
				driver->rx.buff[ driver->rx.count++] = byte;
		        
				driver->rx.fcs = IR_FCS ( INIT_FCS, byte);
				state = INSIDE_FRAME;
			} else 
				printk( "Rx buffer overflow\n");
			break;
		}
		break;
	case LINK_ESCAPE:
		switch ( byte) {
		case BOF:
			/* New frame? */
			DEBUG(4, "New frame?");
			state = BEGIN_FRAME;
			driver->rx.count = 0;
			irport_set_media_busy( driver, TRUE);
			break;
		case CE:
			DEBUG(4, "WARNING: State not defined\n");
			break;
		case EOF:
			/* Abort frame */
			DEBUG(4, "Abort frame (2)\n");
			state = OUTSIDE_FRAME;
			driver->rx.count = 0;
			break;
		default:
			/* 
			 *  Stuffed char, complement bit 5 of byte 
			 *  following CE, IrLAP p.114 
			 */
			byte ^= IR_TRANS;
			if ( driver->rx.count < driver->rx.buffsize)  {
				driver->rx.buff[ driver->rx.count++] = byte;
		        
				driver->rx.fcs = IR_FCS( driver->rx.fcs, byte);
				state = INSIDE_FRAME;
			} else 
				printk( "Rx buffer overflow\n");
			break;
		}
		break;
	case INSIDE_FRAME:
		switch ( byte) {
		case BOF:
			/* New frame? */
			DEBUG(4, "New frame?");
			state = BEGIN_FRAME;
			driver->rx.count = 0;
			irport_set_media_busy( driver, TRUE);
			break;
		case CE:
			/* Stuffed char */
			state = LINK_ESCAPE;
			break;
		case EOF:
			DEBUG(4, "EOF found.\n");
			/* End of frame */
			state = OUTSIDE_FRAME;
			driver->rx.in_frame = FALSE;
			/* 
			 *  Test FCS and deliver frame! 
			 */		        
			if ( driver->rx.fcs == GOOD_FCS) {
				skb = dev_alloc_skb(driver->rx.count);
				if (skb == NULL)  {
					printk( KERN_INFO 
						"async_unwrap: memory squeeze" 
						", dropping frame.\n");
					/* sl->rx_dropped++; */
					return;
				}
				ALLOC_SKB_MAGIC(skb);
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
				skb->free=1;
#endif
				ASSERT( driver->rx.count-2 > 0, return;);
				/*
				 *  Copy data without CRC
				 */
				DEBUG(4, "RECEIVED A FRAME!: Frame len=%d\n", driver->rx.count-2);
				skb_put( skb, driver->rx.count-2);
				memcpy( skb->data, driver->rx.buff, 
					driver->rx.count-2);
				driver->rx.count = 0;
				skb->mac.raw=skb->data;
				skb_queue_tail( &(driver->rx.list), skb);
        
        /* Wake up any readers who are blocking */
        wake_up_interruptible(&driver->readq);

				/*
				 *  Now we have something to chew on so inform upper 
				 *  layers.This will schedule a bottom half routine 
				 *  which is responsible for delivering all 
				 *  received frames.
				 */
				irport_set_media_busy(driver, FALSE);
			} else {
				/* 
				 *  Wrong CRC, discard frame! 
				 */
				DEBUG(0, "Received frame has wrong CRC\n");
				irport_set_media_busy( driver, TRUE); 
			}
			driver->rx.count = 0;
			break;
		default:
			/* Next byte of frame */
			if ( driver->rx.count < driver->rx.buffsize)  {
				driver->rx.buff[ driver->rx.count++] = byte;
			        
				driver->rx.fcs = IR_FCS( driver->rx.fcs, byte);
			} else 
				printk( "Rx buffer overflow\n");
			break;
		}
		break;
	}
	driver->rx.state = state;
}

/* 
 *   Module stuff 
 */

/*
 * Function cleanup_module (void)
 *
 *    
 *
 */
void cleanup_module(void)
{
	DEBUG(4, "IrPORT: cleanup_module!\n");
	irport_cleanup(irport_drv);
}

/*
 * Function init_module (void)
 *
 *    
 *
 */
int init_module(void)
{
  /* 1 - Initialize the driver */
	if (irport_init() < 0) {
			cleanup_module(); 
		return 1;
	}

	return(0);
}

#define MIN_LENGTH 14

inline static int stuff_byte( __u8 byte, __u8 *buf);

/*
 * Function async_wrap (skb, *tx_buff)
 *
 *    Makes a new buffer with wrapping and stuffing, should check that 
 *    we don't get tx buffer overflow.
 */
int irport_async_wrap( struct sk_buff *skb, __u8 *tx_buff, int buffsize) 
{
	__u8 byte;
	__u16 fcs;
	int i, xbofs;
	int n;

	DEBUG(6, "async_wrap()\n"); 
	ASSERT( skb != NULL, return 0;);

	/* Initialize variables */
	fcs = INIT_FCS;
	n = 0;

	if ( skb->len > 2048) {
		DEBUG(0,"async_xmit: Warning size=%d of sk_buff to big!\n", 
		       (int) skb->len);
		IS_SKB( skb, return 0;);
		FREE_SKB_MAGIC( skb);
		DEV_KFREE_SKB( skb, FREE_WRITE);
		return 0;
	}       

	/* Send  XBOF's for required min. turn time */
	xbofs = irport_drv->baud * 0x01 / 10000000;
	for ( i=0; i<xbofs; i++) {
		tx_buff[n++] = XBOF; /* XBOF */
	}

	/* Send negotiated additional BOFs */
  /*
	xbofs = irlap->bofs_count;
	for ( i=0; i<xbofs; i++) {
		tx_buff[n++] = BOF; 
	}
  */
        
	/* Start of packet character BOF */
	tx_buff[n++] = BOF;

	/* Insert frame and calc CRC */
	for( i=0; i < skb->len; i++) {
		byte = skb->data[i];
	        
		/*
		 *  Check for the possibility of tx buffer overflow. We use
		 *  bufsize-5 since the maximum number of bytes that can be 
		 *  transmitted after this point is 5.
		 */
		if ( n > buffsize-5) {
			printk( KERN_WARNING 
				"IrDA Wrapper: TX-buffer overflow!\n");
			return n;
		}
		n+=stuff_byte( byte, tx_buff+n);
		fcs = IR_FCS( fcs, byte);
	}
 
	/* Insert CRC ( LSB first) */
	fcs = ~fcs;
	n += stuff_byte( (__u8) fcs & 0xff, tx_buff+n);
	n += stuff_byte( (__u8) (fcs >> 8) & 0xff, tx_buff+n);

	tx_buff[n++] = EOF;
        
	DEBUG(6, "async_wrap() -->\n");
        
	return n;
}
 
/*
 * Function stuff_byte (byte, buf)
 *
 *    Byte stuff one single byte and put the result in buffer pointed to by
 *    buf. The buffer must at all times be able to have two bytes inserted.
 * 
 */
inline static int stuff_byte( __u8 byte, __u8 *buf) 
{
	switch ( byte) {
	case BOF:
	case EOF:
	case CE:
		/* Insert transparently coded */
		buf[0] = CE;		       /* Send link escape */
		buf[1] = byte^IR_TRANS;    /* Complement bit 5 */
		return 2;
		/* break; */
	default:
		 /* Non-special value, no transparency required */
		buf[0] = byte;
		return 1;
		/* break; */
	}
}
